/*
 * fugit.c: trivial time synchronisation over the network
 *
 * Copyright (c) 2000 Chris Lightfoot. All rights reserved.
 *
 * $Log: fugit.c,v $
 * Revision 1.3  2000/08/27 00:49:12  chris
 * Minor changes.
 *
 * Revision 1.2  2000/08/27 00:47:14  chris
 * Documentation, options.
 *
 * Revision 1.1  2000/08/27 00:29:57  chris
 * Initial revision
 *
 *
 */

static char rcsid[] = "$Id: fugit.c,v 1.3 2000/08/27 00:49:12 chris Exp chris $";

#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/timex.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <math.h>

#include "fugit.h"

/*
 * FUGIT -- simple time synchronisation, to reasonable (ms) accuracy taking
 *          account of network delays.
 * 
 * The plan is very simple:
 *
 * The client sends--
 *
 * +---+-----------+------------+-------------+
 * |'F'| id (long) | sec (long) | usec (long) |
 * +---+-----------+------------+-------------+
 *
 * The server returns--
 * 
 * +---+-----------------+-------------------+-----------+------------+-------------+
 * |'F'| local sec(long) | local usec (long) | id (long) | sec (long) | usec (long) |
 * +---+-----------------+-------------------+-----------+------------+-------------+
 *
 * the latter three fields being copied from the received packet.
 *
 * If running as root, it then adjusts the time using an offset which is
 * calculated by:
 *
 * time at remote server now = time at remote server on packet transmission
 *                             + difference between local time of packet
 *                               transmission and reception
 *
 * Then, we use adjtimex(2) to adjust the time, using the simple one-shot
 * semantics rather than the complicated ones. If the offset is greated than
 * the 131071 specified, we simply use settimeofday to force it to be correct,
 * which sucks badly but works.
 * 
 * Use by running the server ("fugit -s") on a machine whose clock you know to
 * be correct -- this is important, since if the time on the remote machine is
 * very different to that on the local machine, it will force the local
 * machine's clock straight to the remote one -- and then run "fugit <host>"
 * against the server.
 * 
 */

int verbose;
 
void usage() {
	fprintf(stderr, "fugit: simple time synchronisation with a remote known-good clock\n"
			"\n"
			"Usage: fugit [-v|--verbose] ([-s|--server] | host)\n"
			"\n"
			"Run in server mode on a machine with a known-good clock; on a client,\n"
			"it will attempt to synchronise the local clock with the server. Verbose\n"
			"mode will give you all the gory details.\n"
			"\n"
			"%s\n", rcsid);
	exit(1);
}

void run_client(int sck, char *host) {
	struct sockaddr_in sin;
	unsigned char client_buffer[CLIENT_PACKET_LEN], server_buffer[SERVER_PACKET_LEN];
	struct timeval tv;
	long num;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(SERVER_PORT);

	if (!inet_aton(host, &(sin.sin_addr))) {
		struct hostent *he;
		he = gethostbyname(host);
		if (!he) {
			fprintf(stderr, "fugit: gethostbyname: %s\n", strerror(h_errno));
			exit(1);
		}
		memcpy(&(sin.sin_addr), he->h_addr, sizeof(sin.sin_addr));
	}
	
	if (connect(sck, &sin, sizeof(sin)) == -1) {
		perror("fugit: connect");
		exit(1);
	}

	for (num = 0;; num++) {
		int i, j = sizeof(sin);
		char *p = client_buffer;

		memset(client_buffer, 0, sizeof(client_buffer));
		*p++ = PACKET_ID;
		memcpy(p, &num, sizeof(long));		p += sizeof(long);

		/* transmit local time */
		gettimeofday(&tv, NULL);
		tv.tv_sec = htonl(tv.tv_sec);
		tv.tv_usec = htonl(tv.tv_usec);
		memcpy(p, &(tv.tv_sec), sizeof(long));	p += sizeof(long);
		memcpy(p, &(tv.tv_usec), sizeof(long));	p += sizeof(long);
	
		i = sendto(sck, client_buffer, sizeof(client_buffer), 0, &sin, sizeof(sin));
		
		if (i < sizeof(client_buffer) && i > 0)
			fprintf(stderr, "fugit: transmitted short packet (%d bytes) (oops)\n", i);
		else if (i == -1) {
			perror("fugit: sendto");
			exit(1);
		}

		/* should select(2) here, since otherwise we may get desynchronised if a packet is dropped */
		i = recvfrom(sck, server_buffer, sizeof(server_buffer), 0, &sin, &j);

		if (i == sizeof(server_buffer)) {
			if (*server_buffer != PACKET_ID)
				fprintf(stderr, "fugit: bad packet (wrong lead octet) received from %s", inet_ntoa(sin.sin_addr));
			else {
				char *p = server_buffer + 1;
				double here_now, here_then, there_then, there_now;
				long a;

				memcpy(&a, p, sizeof(long));	p += sizeof(long);
				there_then = ntohl(a);
				memcpy(&a, p, sizeof(long));	p += sizeof(long);
				there_then += ntohl(a) * 1E-6;

				memcpy(&a, p, sizeof(long));	p += sizeof(long);
				if (a == num) {
					gettimeofday(&tv, NULL);
					here_now = tv.tv_sec + tv.tv_usec * 1E-6;
					
					memcpy(&a, p, sizeof(long));	p += sizeof(long);
					here_then = ntohl(a);
					memcpy(&a, p, sizeof(long));	p += sizeof(long);
					here_then += ntohl(a) * 1E-6;

					there_now = there_then + (here_now - here_then);

					if (verbose)
						printf(" delay: %lf\n"
							"offset: %lf\n"
							"\t here then: %lf\n"
							"\t  here now: %lf\n"
							"\tthere then: %lf\n"
							"\t there now: %lf (estimated)\n",
							here_now - here_then,
							there_now - here_now,
							here_then,
							here_now,
							there_then,
							there_now);

					/* now use adjtimex to fix the problem */
					if (!getuid()) {
						struct timex adj;

						memset(&adj, 0, sizeof(adj));
						if (adjtimex(&adj) == -1) perror("fugit: adjtimex(get)");
						else {
							adj.modes = ADJ_OFFSET_SINGLESHOT;
							adj.constant = 16;
							adj.offset = (int)((there_now - here_now) * 1E6) / 4;
							if (verbose) printf("adj.offset = %d\n", adj.offset);

							if (abs(adj.offset) > 131070) {
								struct timeval tv2;
								tv2.tv_sec = (long)there_now;
								tv2.tv_usec = (long)((there_now - floor(there_now)) * 1E6);
								if (settimeofday(&tv2, NULL) == -1) perror("fugit: settimeofday");
								else fprintf(stderr, "fugit: forced settimeofday adjustment from large offset\n");
							} else if (adjtimex(&adj) == -1) perror("fugit: adjtimex(set)");
						}
					}

				} else fprintf(stderr, "fugit: possible packet synchronisation problem\n");
			}
		} else if (i >= 0)
			fprintf(stderr, "fugit: short packet (%d bytes) received from %s\n",
				i, inet_ntoa(sin.sin_addr));
		else if (i == -1) {
			perror("fugit: recvfrom");
			exit(1);
		}

		sleep(TRANSMIT_DELAY);
	}
}

void run_server(int sck) {
	struct sockaddr_in sin;
	unsigned char client_buffer[CLIENT_PACKET_LEN], server_buffer[SERVER_PACKET_LEN];
	struct timeval tv;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(SERVER_PORT);
	
	if (bind(sck, &sin, sizeof(sin)) == -1) {
		perror("fugit: bind");
		exit(1);
	}

	for (;;) {
		int i, j = sizeof(sin);

		i = recvfrom(sck, client_buffer, sizeof(client_buffer), 0, &sin, &j);

		if (i == sizeof(client_buffer)) {
			if (*client_buffer != PACKET_ID)
				fprintf(stderr, "fugit: bad packet (wrong lead octet) received from %s", inet_ntoa(sin.sin_addr));
			else {
				char *p = server_buffer;
				
				/* construct reply packet */
				memset(server_buffer, 0, sizeof(server_buffer));
				*p++ = PACKET_ID;

				gettimeofday(&tv, NULL);
				tv.tv_sec = htonl(tv.tv_sec);
				tv.tv_usec = htonl(tv.tv_usec);
				memcpy(p, &(tv.tv_sec), sizeof(long));		p += sizeof(long);
				memcpy(p, &(tv.tv_usec), sizeof(long));		p += sizeof(long);

				/* return the client's time to them */
				memcpy(p, client_buffer + 1, 3 * sizeof(long));

				sendto(sck, server_buffer, sizeof(server_buffer), 0, &sin, sizeof(sin));

				fprintf(stderr, "fugit: transmitted reply to %s\n", inet_ntoa(sin.sin_addr));
			}
		} else if (i >= 0)
			fprintf(stderr, "fugit: short packet (%d bytes) received from %s\n",
				i, inet_ntoa(sin.sin_addr));
		else if (i == -1) perror("fugit: recvfrom");
	}
}

int main(int argc, char **argv) {
	int sck;
	char **p;
	char *host = NULL;
	int server = 0;

	for (p = argv + 1; p < argv + argc; p++)
		if (!strcmp(*p, "--server") || !strcmp(*p, "-s")) {
			server = 1;
			if (host) usage();
		} else if (!strcmp(*p, "--verbose") || !strcmp(*p, "-v")) verbose = 1;
		else if (!strcmp(*p, "--help") || !strcmp(*p, "-h")) usage();
		else if (host || server) usage();
		else host = *p;

	if (!server && !host) usage();
	
	if (getuid()) {
		fprintf(stderr, "fugit: you are not root, so will run in read-only, verbose mode\n");
		verbose = 1;
	}

	sck = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sck == -1) {
		perror("fugit: socket");
	}

	if (server) run_server(sck);
	else run_client(sck, host);
}
